from dataclasses import dataclass
from typing import Callable, Generic, TypeVar

from .result import ParseError, Failure
from .transformer import Transformer
from .transformers import (
    EndTransform,
    JustTransform,
    RegexTransform,
    AndTransform,
    OrTransform,
    ListTransform,
    SepListTransform,
    OptionTransform,
    MapTransform,
    ValueTransform
)


def regex(pattern: str):
    return Parser(RegexTransform(pattern))


def just(word: str):
    return Parser(JustTransform(word))


def end():
    return Parser(EndTransform())


P = TypeVar("P")
T = TypeVar("T")
O = TypeVar("O")
@dataclass
class Parser(Generic[P]):
    inner: "Transformer[P]"

    def parse(self, stream: str) -> P:
        parser = self.and_(end()).map(lambda value : value[0])
        parsed = parser.inner.parse(stream, 0)
        if isinstance(parsed, Failure):
            raise ParseError(parsed, stream)
        return parsed.value

    def and_(self, other: "Parser[T]"):
        return Parser(AndTransform(self.inner, other.inner))

    def or_(self, other: "Parser[T]"):
        return Parser(OrTransform(self.inner, other.inner))

    def or_not(self):
        return Parser(OptionTransform(self.inner))

    def repeat(self):
        return Parser(ListTransform(self.inner))

    def map(self, transform: Callable[[P], O]):
        return Parser(MapTransform(self.inner, transform))

    def set(self, value: T) -> "Parser[T]":
        return Parser(ValueTransform(self.inner, value))

    def sep_by(self, sep: "Parser[T]"):
        return Parser(SepListTransform(self.inner, sep.inner))

    # |
    def __or__(self, other: "Parser[T]"):
        return self.or_(other)

    # &
    def __and__(self, other: "Parser[T]"):
        return self.and_(other)

    # >>
    def __rshift__(self, other: "Parser[T]"):
        return self.and_(other).map(lambda v: v[1])

    # <<
    def __lshift__(self, other: "Parser[T]"):
        return self.and_(other).map(lambda v: v[0])
